import sys

import pandas as pd
import numpy as np
from tabulate import tabulate


class RegressionExplanation:
    """Class provides access to prediction explanation data.
    """

    def __init__(
        self,
        pred_value,
        feature_indexes,
        feature_values,
        contributions,
        intercept,
        columns=None,
        classes=None,
    ):
        self.pred_value = pred_value
        self.feature_indexes = feature_indexes
        self.feature_values = feature_values
        self.contributions = contributions
        self.columns = columns
        self.intercept = intercept
        self.digits = 4
        if columns is not None:
            self._columns = columns
        else:
            self._columns = [str(v) for v in range(len(feature_values))]
        if classes is None:
            self._classes = [str(i) for i in range(len(intercept))]
        else:
            self._classes = classes

    def _build_df(self):
        feature_names = [
            idx_to_name(idx, self._columns) for idx in self.feature_indexes
        ]
        feature_names = ['intercept'] + feature_names + ['PREDICTION']
        feature_values = [None] + self.feature_values + [None]
        contrib = [self.intercept] + self.contributions + [self.pred_value]

        data = {
            'Feature Name': feature_names,
            'Feature Value': feature_values,
            'Contributions': contrib,
        }
        df = pd.DataFrame(data)
        return df

    def print(self, file=sys.stdout, flush=False):
        df = self._build_df()
        table = tabulate(df, tablefmt='psql', headers='keys')
        print(table, file=file, flush=flush)


class ClassificationExplanation(RegressionExplanation):
    def _build_df(self):
        feature_names = [
            idx_to_name(idx, self._columns) for idx in self.feature_indexes
        ]
        feature_names = ['intercept'] + feature_names + ['PREDICTION']
        feature_values = [None] + self.feature_values + [None]

        data = {'Feature Name': feature_names, 'Feature Value': feature_values}
        contrib_columns = [
            f'Contribution:{c}' for c in enumerate(self._classes)
        ]
        contrib_val = np.concatenate(
            ([self.intercept], self.contributions, [self.pred_value])
        )
        data.update(zip(contrib_columns, contrib_val.T))
        df = pd.DataFrame(data)
        return df


def idx_to_name(idx, columns):
    if isinstance(idx, int):
        n = columns[idx]
    else:
        n = tuple(columns[i] for i in idx)
    return n



import numpy as np

from .explanation import ClassificationExplanation, RegressionExplanation
from .utils import (
    normalize_array,
    to_matrix,
    magnituge,
    features_groups,
    feature_group_values,
)


class RegressionExplainer:

    exp_class = RegressionExplanation

    def __init__(self, clf):
        self._clf = clf
        self._data = None
        self._columns = None
        self._baseline = None

    def fit(self, data, columns=None, classes=None):
        self._data = data
        if columns is None:
            columns = list(range(data.shape[1]))

        self._columns = columns
        self._classes = classes or [0]
        self._baseline = self._mean_predict(data)

    def explain(self, row):
        instance = to_matrix(row)
        instance = normalize_array(instance)
        path = self._compute_explanation_path(instance)
        pred_value = self._mean_predict(instance)
        feature_indexes, featrue_values, contrib = self._explain_path(
            path, instance
        )

        return self.exp_class(
            pred_value,
            feature_indexes,
            featrue_values,
            contrib,
            self._baseline,
            self._columns,
            self._classes,
        )

    def _mean_predict(self, data):
        return self._clf.predict(data).mean(axis=0)

    def _compute_explanation_path(self, instance):
        num_rows, num_features = self._data.shape
        important_variables = {}
        groups = features_groups(num_features)
        for group in groups:
            new_data = np.copy(self._data)
            new_data[:, group] = instance[:, group]
            pred_mean = self._mean_predict(new_data)
            if isinstance(group, int):
                impact = pred_mean - self._baseline
            else:
                impact = (
                    pred_mean
                    - self._baseline
                    - np.sum(important_variables[g] for g in group)
                )
            important_variables[group] = impact

        path = self._sort(important_variables)
        return path

    def _sort(self, important_variables):
        return sorted(important_variables.items(), key=lambda v: -abs(v[1]))

    def _explain_path(self, path, instance):
        _, num_features = self._data.shape
        features = set(range(num_features))
        new_data = np.copy(self._data)
        important_variables = []
        while features:
            group, _ = path.pop(0)
            if not set([group] if isinstance(group, int) else group).issubset(
                features
            ):
                continue
            new_data[:, group] = instance[:, group]
            pred_mean = self._mean_predict(new_data)
            important_variables.append((group, pred_mean))
            features.difference_update(
                set([group] if isinstance(group, int) else group)
            )

        cummulative = [v[1] for v in important_variables]
        feature_indexes = [v[0] for v in important_variables]
        contrib = np.diff(np.array([self._baseline] + cummulative), axis=0)
        featrue_values = feature_group_values(feature_indexes, instance)
        return feature_indexes, featrue_values, contrib


class ClassificationExplainer(RegressionExplainer):

    exp_class = ClassificationExplanation

    def _sort(self, important_variables):
        return sorted(important_variables.items(), key=lambda v: -magnituge(v))

    def _mean_predict(self, data):
        return self._clf.predict_proba(data).mean(axis=0)



import os
import re
import sys
from setuptools import setup, find_packages


PY36 = (3, 6, 0)
PY37 = (3, 7, 0)


if sys.version_info < PY36:
    raise RuntimeError('ibreakdown does not support Python earlier than 3.6')


install_requires = ['numpy', 'pandas', 'tabulate']
if sys.version_info < PY37:
    install_requires.append('dataclasses==0.6')


def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


extras_require = {}


def read_version():
    regexp = re.compile(r"^__version__\W*=\W*'([\d.abrc]+)'")
    init_py = os.path.join(
        os.path.dirname(__file__), 'ibreakdown', '__init__.py'
    )
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        else:
            msg = 'Cannot find version in ibreakdown/__init__.py'
            raise RuntimeError(msg)


classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Intended Audience :: Developers',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Operating System :: POSIX',
    'Development Status :: 2 - Pre-Alpha',
]


setup(
    name='ibreakdown',
    version=read_version(),
    description=('ibreakdown - model agnostic explanations with interactions'),
    long_description='\n\n'.join((read('README.rst'), read('CHANGES.txt'))),
    install_requires=install_requires,
    classifiers=classifiers,
    platforms=['POSIX'],
    author='Nikolay Novik',
    author_email='nickolainovik@gmail.com',
    url='https://github.com/jettify/ibreakdown',
    download_url='https://pypi.python.org/pypi/ibreakdown',
    license='Apache 2',
    packages=find_packages(),
    extras_require=extras_require,
    keywords=['ibreakdown', 'model explanation', 'xai'],
    zip_safe=True,
    include_package_data=True,
)



import numpy as np


def multiply_row(row, num_rows):
    return np.repeat(row, repeats=num_rows, axis=0)


def normalize_array(instance):
    return instance.reshape(1, -1)


def magnituge(v):
    return np.linalg.norm(np.array(v[1]), axis=0)


def to_matrix(data):
    return data if not hasattr(data, 'values') else data.values


def features_groups(num_features, exclude_set=None):
    result = list(range(0, num_features))
    for i in range(0, num_features):
        for j in range(i + 1, num_features):
            result.append((i, j))
    return result


def feature_group_values(feature_groups, instance):
    featrue_values = [instance[:, group][0] for group in feature_groups]
    return featrue_values
